JOB_TYPE_WORKER_CONTROL = "worker_control"
WORKER_CONTROL_ACTION_RETIRE = "retire"

# TUBE_DEFAULT_NAME = "default"
# BROKER_HOST_DEFAULT_ADDRESS = "fwq-the_broker"
BROKER_HOST_DEFAULT_PORT = 11300

JOB_START_PERSISTENCE_DELAY_IN_SECS = 1

JOB_STATE_FINAL = "finalized"
JOB_STATE_DONE = "done"
JOB_STATE_NEW = "new"
JOB_STATE_DELAYED = "delayed"
JOB_STATE_RESERVED = "reserved"
JOB_STATE_READY = "ready"
JOB_STATE_BURIED = "buried"

FINAL_JOB_STATES = [JOB_STATE_FINAL, JOB_STATE_BURIED]
PEEK_NEXT_JOB_STATES = [JOB_STATE_READY, JOB_STATE_DELAYED, JOB_STATE_BURIED]

STATS_TYPE_SYSTEM = "system"
STATS_TYPE_TUBE = "tube"
STATS_TYPE_JOB = "job"

BROKER_CONFIG_FROM_JOB_DATA = '_fwq_broker_config'
WORK_Q_CONFIG_FROM_JOB_DATA = '_fwq_work_q_config'
WORKER_CONFIG_FROM_JOB_DATA = '_fwq_worker_config'




from time import time
from datetime import timedelta

import boto3

from gwerks import now
from fwq.logs import Sender


class CloudWatchLogsSender(Sender):
    def __init__(self, log_group, log_stream, batch_timespan_in_minutes=5, max_batch_items=100, max_batch_size_bytes=1048576):

        if max_batch_size_bytes > 1048576:
            raise Exception(f"max_batch_size_bytes too big, must be 1,048,576 or less")

        if max_batch_items > 10000:
            raise Exception(f"max_batch_size too big, must be 10,000 or less")

        fouteen_days_in_minutes = 60 * 24 * 14
        if batch_timespan_in_minutes > fouteen_days_in_minutes:
            raise Exception(f"batch_timespan_in_minutes too big, must be less than {fouteen_days_in_minutes}")

        self._cw_logs = boto3.client('logs')
        self._log_group_name = log_group
        self._log_stream_name = log_stream
        self._batch_timespan_in_minutes = batch_timespan_in_minutes
        self._max_batch_items = max_batch_items
        self._max_batch_size_bytes = max_batch_size_bytes

        self._make_new_batch()

    def __repr__(self):
        return f"<CloudWatchLogsSender log_group:{self._log_group_name} log_stream:{self._log_stream_name}>"

    def send(self, msg):

        kb256 = 256 * 1024
        msg_len = len(msg)

        # truncate messages that are too long
        if msg_len > kb256:
            msg = msg[:kb256]

        # flush if batch is too big
        if self._batch_size_bytes + msg_len + 26 > self._max_batch_size_bytes:
            self.flush()

        # batch am event
        log_event = {
            'timestamp': time(),
            'message': msg
        }
        self._log_event_batch.append(log_event)
        self._batch_size_bytes += msg_len + 26

        # flush if batch is full
        if len(self._log_event_batch) >= self._max_batch_items:
            self.flush()

        # flush if batch is aging out
        if now() > self._batch_start_time + timedelta(minutes=self._batch_timespan_in_minutes):
            self.flush()

    def flush(self):
        resp = self._cw_logs.put_log_events(
            logGroupName=self._log_group_name,
            logStreamName=self._log_stream_name,
            logEvents=self._log_event_batch,
            # entity={
            #     'keyAttributes': {
            #         'string': 'string'
            #     },
            #     'attributes': {
            #         'string': 'string'
            #     }
            # }
        )
        # print(f"{yaml.dump(resp)}")
        self._make_new_batch()

        results = ""
        if "rejectedLogEventsInfo" in resp:
            results += f"Log events were rejected: {resp['rejectedLogEventsInfo']}. "
        if "rejectedEntityInfo" in resp:
            results += f"Entities were rejected: {resp['rejectedEntityInfo']}. "
        if results != "":
            raise Exception(results)

    def _make_new_batch(self):
        self._log_event_batch = []
        self._batch_start_time = now()
        self._batch_size_bytes = 0


import os

from gwerks import emitter, is_dev_environment
from gwerks.docker import DockerApp, DockerContext, DockerNetwork


def parse_broker_str(broker: str):
    broker_s = broker.split(":")
    if len(broker_s) == 1:
        broker_tpl = (broker_s[0], int(Broker.DEFAULT_PORT))
    else:
        broker_tpl = (broker_s[0], int(broker_s[1]))
    return broker_tpl


@emitter()
class Broker(DockerApp):

    DOCKERFILE_BASE = ""
    DOCKERFILE_BASE += "FROM python:bullseye\n"
    DOCKERFILE_BASE += "RUN apt-get update\n"
    DOCKERFILE_BASE += "RUN pip3 install --upgrade pip\n"
    DOCKERFILE_BASE += "RUN pip3 --no-cache-dir install --upgrade awscli\n"
    DOCKERFILE_BASE += "ARG AWS_DEFAULT_REGION\n"
    DOCKERFILE_BASE += "ARG AWS_CONTAINER_CREDENTIALS_RELATIVE_URI\n"
    DOCKERFILE_BASE += "ARG AWS_ACCESS_KEY_ID\n"
    DOCKERFILE_BASE += "ARG AWS_SECRET_ACCESS_KEY\n"
    DOCKERFILE_BASE += "RUN apt-get install beanstalkd\n"

    DOCKERFILE = DOCKERFILE_BASE
    DOCKERFILE += "RUN pip3 install fwq\n"

    DOCKER_IMAGE_NAME = "fwq_broker"

    DEFAULT_NAME = "fwq_broker"
    DEFAULT_PORT = "11300"
    DEFAULT = f"{DEFAULT_NAME}:{DEFAULT_PORT}"

    def __init__(self, app_name=DEFAULT_NAME, context=None):
        super().__init__(app_name, Broker.DOCKER_IMAGE_NAME)

        if not context:
            context = DockerContext({
                "dockerfile_str": Broker.DOCKERFILE,
            })
            if is_dev_environment():
                context.set('pass_through_cloud_creds', "aws")
        self._context = context

    # --------------------------------------------------------------------------- #
    # delegates to run
    def start(self, port_spec: str = DEFAULT_PORT, network: DockerNetwork = None,
              map_to_host_data_volume: str = None, max_message_size: int = 65535, log_group_prefix: str = None):

        self.stop()
        self.remove()

        self._context.build(self.get_image_name())

        port_mappings = []
        if port_spec == "^":
            host_port_open = True
            port = int(Broker.DEFAULT_PORT)
            port_mappings.append((port, port))
            print(f"port_mappings: {port_mappings}")
        elif port_spec.endswith("^"):
            host_port_open = True
            port = int(port_spec[:-1])
            port_mappings.append((port, port))
            print(f"port_mappings: {port_mappings}")
        else:
            host_port_open = False
            port = int(port_spec)

        volume_mappings = []
        if map_to_host_data_volume:
            if not os.path.exists(map_to_host_data_volume):
                print(f"makedirs: {map_to_host_data_volume}")
                os.makedirs(map_to_host_data_volume)
            volume_mappings.append((map_to_host_data_volume, "/data"))
            print(f"volume_mappings: {volume_mappings}")

        # set up the log group
        if not log_group_prefix:
            log_group_prefix = ""
        log_group_prefix += f"/{self.get_app_name()}_{port}"
        print(f"log_group_prefix: {log_group_prefix}")

        app_cmd = f"beanstalkd -l 0.0.0.0 -p {port} -b /data -V -z {max_message_size}"
        print(f"app_cmd: {app_cmd}")

        self.run(app_cmd, network=network, port_mappings=port_mappings, volume_mappings=volume_mappings,
                 log_group_prefix=log_group_prefix)
        print(f"Ready to receive:")
        if host_port_open:
            print(f"broker (external): {self.get_host_name()}:{port}")
        print(f"broker (internal): {self.get_app_name()}:{port}")

    # --------------------------------------------------------------------------- #
    # delegates to stop, remove
    def halt(self):
        self.stop()
        self.remove()
        print(f"Stopped!")



import json

from greenstalk import Client, NotFoundError

from fwq.broker import Broker, parse_broker_str
from fwq.constants import JOB_START_PERSISTENCE_DELAY_IN_SECS, JOB_STATE_READY, JOB_STATE_BURIED, JOB_STATE_DELAYED


class WorkQ:
    def __init__(self, for_app: str, broker: str = Broker.DEFAULT):
        self._for_app = for_app
        self._broker = broker

        broker_tpl = parse_broker_str(broker)
        self._client = Client(broker_tpl, use=for_app, watch=[for_app])

        print(f"WorkQ {self._for_app} is connected to {self._broker}")

    def kick(self, job_id=None, how_many=None):
        number_kicked = 0
        try:
            if job_id is not None:
                self._client.kick_job(int(job_id))
                number_kicked += 1
            elif how_many is not None:
                number_kicked = self._client.kick(int(how_many))
        except NotFoundError:
            if job_id is not None:
                print(f"job {job_id} not found")
            else:
                print(f"job not found")
        return number_kicked

    def put(self, job_type, job_data=None, job_name=None, priority=65536, delay=0, ttr=60):

        if job_name is None:
            job_name = _make_job_name(job_type, job_data)

        if job_data is None:
            job_data = {}
        # job_data['_fwq_work_q_config'] = json.dumps(self._config)  # TODO fix

        job_body = json.dumps({
            "type": job_type,
            "data": job_data,
            "name": job_name
        })
        job_id = self._client.put(job_body, priority=priority, delay=delay + JOB_START_PERSISTENCE_DELAY_IN_SECS,
                                  ttr=ttr)
        print(f"NQ Job [{job_id}] - {job_name} - [priority={priority}, delay={delay}, ttr={ttr}]")
        return job_id

    def peek(self, job_id=None, next_in_state=JOB_STATE_READY):
        job = None
        try:
            if job_id is not None:
                job = self._client.peek(int(job_id))
            elif next_in_state == JOB_STATE_READY:
                job = self._client.peek_ready()
            elif next_in_state == JOB_STATE_BURIED:
                job = self._client.peek_buried()
            elif next_in_state == JOB_STATE_DELAYED:
                job = self._client.peek_delayed()
        except NotFoundError:
            if job_id is not None:
                print(f"job {job_id} not found")
            else:
                print(f"job not found")
        return job

    def purge(self, at_most, in_state=JOB_STATE_READY):
        purged_count = 0
        while purged_count < int(at_most):
            job = self.peek(next_in_state=in_state)
            if job is None:
                break
            try:
                self._client.delete(job)
                purged_count += 1
                print(f"{in_state} job {job.id} deleted")
            except NotFoundError:
                print(f"job {job.id} not found")

    def touch(self, job_id):
        job = self._client.peek(int(job_id))
        self._client.touch(job)

    # def use(self, a_tube_name):
    #     self._use_tube_name = a_tube_name
    #     use_tube = make_tube_name(self._for_app, self._use_tube_name)
    #     self._client().use(use_tube)

    # def all_jobs_status(self, for_state=None):
    #     results = []
    #     j_rows = self._db.select_active_jobs(self._beanstalk_server_id)
    #     if j_rows is not None:
    #         for j_row in j_rows:
    #             job_status = self.job_status(make_sj_id(self._beanstalk_server_id, j_row['job_id']))
    #             if job_status is not None:
    #                 if for_state is None or job_status['job_state'] == for_state:
    #                     results.append(job_status)

        # return results

    def stats(self, for_job_id=None):
        job_stats = None
        if for_job_id is not None:
            try:
                job_stats = self._client.stats_job(int(for_job_id))
            except NotFoundError:
                print(f"job {for_job_id} not found")

        tube_stats = self._client.stats_tube(self._for_app)

        sys_stats = self._client.stats()

        return sys_stats, tube_stats, job_stats

    def stop(self):
        if self._client is not None:
            self._client.close()
            self._client = None
            # self._broker = None

    # def _client(self) -> Client:
    #     if self._bs_client is None:
    #         self._bs_client, self._broker = get_beanstalk_client(self._beanstalk_host, self._for_app)
    #         # the_tube = make_tube_name(self._for_app, self._use_tube_name)
    #         print(f"WorkQ {self._for_app} is connected to {self._beanstalk_host}")
    #     return self._bs_client


def _make_job_name(job_type: str, job_data):
    module = job_type
    try:
        pkg, module = job_type.rsplit(".", 1)
    except Exception:
        pass
    return f"{module}-{abs(hash(json.dumps(job_data))) % (10 ** 8)}"



from fwq.broker import Broker
from fwq.worker import Worker
from fwq.work_q import WorkQ
from fwq.constants import JOB_STATE_READY

from gwerks.docker import DockerNetwork, DockerContext


def start(name=Broker.DEFAULT_NAME, port=Broker.DEFAULT_PORT, net_name=None, net_driver=DockerNetwork.DEFAULT_DRIVER,
          host_volume=None, max_message_size=65535, log_group_prefix=None, context_yaml_file=None):
    network = None
    if net_name:
        network = DockerNetwork(net_name, driver=net_driver)

    context = None
    if context_yaml_file:
        context = DockerContext(from_yaml_file=context_yaml_file)

    the_broker = Broker(app_name=name, context=context)

    the_broker.start(port_spec=port, network=network, map_to_host_data_volume=host_volume,
                     max_message_size=max_message_size, log_group_prefix=log_group_prefix)


def stop(name=Broker.DEFAULT_NAME):
    the_broker = Broker(app_name=name)
    the_broker.halt()


def do_next_job(for_app, the_broker=f"{Broker.DEFAULT_NAME}:{Broker.DEFAULT_PORT}", wait_secs=60):
    the_worker = Worker(for_app, the_broker)
    the_worker.do_job(wait_for_job_secs=wait_secs)


def do_jobs(for_app, the_broker=f"{Broker.DEFAULT_NAME}:{Broker.DEFAULT_PORT}", wait_secs=60):
    the_worker = Worker(for_app, the_broker)
    the_worker.do_jobs(wait_for_job_secs=wait_secs)


def kick_job(for_app, the_broker=f"{Broker.DEFAULT_NAME}:{Broker.DEFAULT_PORT}", job_id=None, kick_count=None):
    the_work_q = WorkQ(for_app, the_broker)
    the_work_q.kick(job_id=job_id, how_many=kick_count)


def stats(for_app, the_broker=f"{Broker.DEFAULT_NAME}:{Broker.DEFAULT_PORT}", job_id=None):
    the_work_q = WorkQ(for_app, the_broker)
    system_stats, tube_stats, job_stats = the_work_q.stats(for_job_id=job_id)
    return system_stats, tube_stats, job_stats


def nq(for_app, job_type, the_broker=f"{Broker.DEFAULT_NAME}:{Broker.DEFAULT_PORT}", job_data=None, job_name=None,
       priority=65536, delay_secs=0, ttr=60):
    the_work_q = WorkQ(for_app, the_broker)
    the_work_q.put(job_type, job_data, job_name, priority, delay_secs, ttr)


def purge(for_app, the_broker=f"{Broker.DEFAULT_NAME}:{Broker.DEFAULT_PORT}", at_most=None, in_state=JOB_STATE_READY):
    the_work_q = WorkQ(for_app, the_broker)
    the_work_q.purge(at_most=at_most, in_state=in_state)


def peek(for_app, the_broker=f"{Broker.DEFAULT_NAME}:{Broker.DEFAULT_PORT}", job_id=None, next_in_state=JOB_STATE_READY):
    the_work_q = WorkQ(for_app, the_broker)
    return the_work_q.peek(job_id=job_id, next_in_state=next_in_state)



"""'Context manager class to capture job output'"""

import inspect
import sys
import traceback
from typing import Protocol


class Sender(Protocol):
    def send(self, msg):
        ...

    def flush(self):
        ...


class JobLogger:
    def __init__(self, broker_id, worker_id):
        self._broker_id = broker_id
        self._worker_id = worker_id
        self._job_id = None
        self._stdout = sys.stdout
        self._streams: list[Sender] = []

    def register_logstream(self, logstream: Sender):
        self._streams.append(logstream)

    def change_jobs(self, new_job_id):
        self._job_id = new_job_id

    def __enter__(self):
        sys.stdout = self

    def write(self, msg):
        try:
            msg = msg.strip()
            if len(msg) > 0:
                msg_source = inspect.stack()[1].function
                msg = self._format_msg(f"[{msg_source}] {msg}")
                self._stdout.write(msg)
                for ls in self._streams:
                    try:
                        ls.send(msg)
                    except Exception as e:
                        err_msg = self._format_msg(f"[{msg_source}] {ls} ERROR: {e}")
                        self._stdout.write(err_msg)

                # self.flush()
        except Exception as e:
            self._stdout.write(self._format_msg(f"ERROR: {e}"))

    def _format_msg(self, msg):
        return f"JOB [{self._broker_id}]-[{self._worker_id}]-[{self._job_id}]: {msg}\n"

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is not None:
            self.write(traceback.format_exc())
        sys.stdout = self._stdout

    def flush(self):
        for ls in self._streams:
            try:
                ls.flush()
            except Exception as e:
                err_msg = self._format_msg(f"{ls} ERROR: {e}")
                self._stdout.write(err_msg)
        self._stdout.flush()



"""'Fast Work Queue'"""
from gwerks.packaging import get_version
__version__ = get_version(__file__)


# TODO job dependencies
# TODO worker control may need to listen to a separate control channel, test it regardless
# TODO alerting hook and Slack implementation
# TODO baseline cli config that has runtime_env setting (and region?)

"""


from gwerks.docker import DockerContext
from fwq.the_broker import Broker
DockerContext({
    "dockerfile_str": Broker.DOCKERFILE_BASE + "COPY gwerks.tar.gz .\nCOPY fwq.tar.gz .\nRUN pip3 install gwerks.tar.gz\nRUN pip3 install fwq.tar.gz\n",
    "files": [["../gwerks/dist/gwerks-25.2.13.tar.gz", "gwerks.tar.gz"], ["dist/fwq-0.0.0.tar.gz", "fwq.tar.gz"]]
}).to_yaml_file("./fwq_broker_docker_context-Dev.example.yaml")

DockerContext(from_yaml_file="./fwq_broker_docker_context-Dev.example.yaml").build("fwq_dev_image", no_cache=False)

"""


# --------------------------------------------------------------------------- #





import time


def test_job(job_info):

    job_id = job_info['job_id']
    job_data = job_info['job_data']

    print(f"test_job RUNNING")
    print(f"job_id: {job_id}")
    print(f"job_data: {job_data}")

    print(f"test_job SLEEPING")
    time.sleep(5)

    print(f"I'm awake!")

    return "{'some': 'info'}"
